 ! Copyright 2019 Thomas A. Manz. 
 ! Software License: CC BY - Creative Commons Attribution

 MODULE module_compute_C6tot_using_wp_table
 
 USE, INTRINSIC :: iso_fortran_env
!$ USE omp_lib

 IMPLICIT NONE 

 CONTAINS
 
 FUNCTION C6tot(alpha_values,wp_values,Natoms,Num_lookup)
 !----------------------------------------------------------------------
 ! Computes C6 total using a wp lookup table
 !----------------------------------------------------------------------
 INTEGER,INTENT(IN) :: Natoms,Num_lookup
 INTEGER,PARAMETER :: dp = REAL64
 INTEGER,PARAMETER :: qp = REAL128
 REAL(kind=dp),INTENT(IN) :: alpha_values(Natoms),wp_values(Natoms)
 REAL(kind=qp) :: C6tot,temp_real_qp,alpha_table(Num_lookup)
 REAL(kind=dp) :: log_wp_table(Num_lookup),wp_table(Num_lookup),temp_real_1,temp_real_2,max_log_wp, min_log_wp,interval_log_wp
 INTEGER :: i,atom1,index_1,count
 LOGICAL :: wp_lookup_used(Num_lookup)

 wp_lookup_used = .FALSE.

 !Note that adding +/- 0.01_dp amount makes interval_log_wp non-zero even when min(wp) = max(wp)
 !It also ensures that the calculated array indes does not round outside of 1 or Num_lookup
 min_log_wp = log(MINVAL(wp_values)) - 0.01_dp
 max_log_wp = log(MAXVAL(wp_values)) + 0.01_dp
 interval_log_wp = (max_log_wp - min_log_wp)/(Num_lookup - 1.0_dp)

 ! Construct the log_wp_table
 DO i = 1,Num_lookup
     log_wp_table(i) = min_log_wp + (i-1)*interval_log_wp
     wp_table(i) = exp(log_wp_table(i))
 END DO

 ! Construct the alpha_table
 alpha_table = 0.0_dp
 DO atom1=1,Natoms
    index_1 = FLOOR( (log(wp_values(atom1)) - min_log_wp)/interval_log_wp + 1.0_dp )
    temp_real_1 = (wp_values(atom1) - wp_table(index_1))/(wp_table(index_1 + 1) - wp_table(index_1)) 
    alpha_table(index_1) = alpha_table(index_1) + (1.0_dp - temp_real_1)*alpha_values(atom1)
    alpha_table(index_1 + 1) = alpha_table(index_1 + 1) + temp_real_1*alpha_values(atom1) 
    wp_lookup_used(index_1) = .TRUE.  
    wp_lookup_used(index_1 + 1) = .TRUE.    
 END DO

 ! Calculate C6tot
 C6tot = 0.0_dp 
!$omp parallel do default(none) &
!$omp private(i,atom1,temp_real_qp) &
!$omp shared(Num_lookup,Natoms,wp_lookup_used,alpha_values,alpha_table,wp_values,wp_table) &
!$omp reduction(+:C6tot) &
!$omp schedule(dynamic,250) 
 DO i = 1,Num_lookup
    IF (wp_lookup_used(i)) THEN
      temp_real_qp = 0.0_dp
      DO atom1 = 1,Natoms
         temp_real_qp = temp_real_qp + alpha_values(atom1)*(wp_values(atom1)*wp_table(i))/(wp_values(atom1) + wp_table(i))    
      END DO 
      C6tot = C6tot + 1.5_dp*alpha_table(i)*temp_real_qp
    END IF
 END DO    
!$omp end parallel do

 END FUNCTION C6tot

 END MODULE module_compute_C6tot_using_wp_table
