 ! Copyright 2019 Thomas A. Manz. 
 ! Software License: CC BY - Creative Commons Attribution

 MODULE module_initialize_atom_image_pair_matrices
 
 USE precise
 USE common_variable_declarations
 USE module_screening_functions
 USE matrix_operations
!$ USE omp_lib

 IMPLICIT NONE
 
 CONTAINS
 
 SUBROUTINE initialize_atom_image_pair_matrices()
 !----------------------------------------------------------------------
 
 INTEGER :: temp_count_large,temp_count_small,repeata,repeatb,repeatc,ref_regions_A,ref_regions_B,ref_regions_C,num_periodic_dir,&
 temp_count,prior_temp_count,num_region_pairs,index_1,index_2,index_3,index_4,prior_num_region_pairs,region_a,region_b,region_c,&
 index_k,start_k,end_k,index_j,start_j,end_j,index_i,start_i,end_i
 INTEGER(kind=8),ALLOCATABLE,DIMENSION(:) :: pair_count_small, pair_count_large
 INTEGER(kind=8),ALLOCATABLE,DIMENSION(:,:) :: region_pairs,region_pair_images,sorted_atoms
 INTEGER(kind=8),ALLOCATABLE,DIMENSION(:,:,:) :: ref_region_count,temp_region_count
 INTEGER(kind=8),ALLOCATABLE,DIMENSION(:,:,:,:) :: region_start_end
 REAL(kind=dp) :: temp1,temp2,temp5,temp4,a_dot_b,a_dot_c,b_dot_c,cos_a_b,cos_a_c,cos_b_c,sin_a_b,&
 sin_a_c,sin_b_c,X1,Y1,Z1,X2,Y2,Z2,temp6,length_a,length_b,length_c,normalization,delta_a,delta_b,delta_c,temp_X2,temp_Y2,temp_Z2,&
 unit_cell_volume,longest_body_diag,dist_squared,dist_squared_cutoff
 REAL(kind=dp) :: basis_vector_a(3),basis_vector_b(3),basis_vector_c(3),max_coords(3),min_coords(3),&
 xyz(3),origin(3),basis_vectors(3,3),inverse_vectors(3,3),pre_basis_vector_a(3),pre_basis_vector_b(3),pre_basis_vector_c(3)
 REAL(kind=dp),ALLOCATABLE,DIMENSION(:,:) :: sorted_coords,inverse_coords
 LOGICAL :: active_pair_large,active_pair_small,active_region
 
 ALLOCATE(pair_count_small(natoms))
 ALLOCATE(pair_count_large(natoms))
 ALLOCATE(sorted_coords(4,natoms))
 ALLOCATE(sorted_atoms(4,natoms))
 ALLOCATE(inverse_coords(3,natoms))

 WRITE(output_FID,*)'Initializing the atom_image_pair_matrices ...'
 WRITE(output_FID,*)' '
 FLUSH(output_FID)
 
 ! ==========================================================================
 ! STEP # 1: Set up the basis vectors and unit cell parallelepiped
 ! ==========================================================================

 ! Set up the three basis vectors for a parallelepiped enclosing the unit cell or molecule
 basis_vector_a = 0.0_dp
 basis_vector_b = 0.0_dp
 basis_vector_c = 0.0_dp
 pre_basis_vector_a = 0.0_dp
 pre_basis_vector_b = 0.0_dp
 pre_basis_vector_c = 0.0_dp

 num_periodic_dir = 0
 IF (periodicA) num_periodic_dir = num_periodic_dir + 1
 IF (periodicB) num_periodic_dir = num_periodic_dir + 1
 IF (periodicC) num_periodic_dir = num_periodic_dir + 1
 IF (num_periodic_dir .EQ. 3) THEN
    !
    WRITE(output_FID,*) ' '
    WRITE(output_FID,*) 'Periodic along directions A, B, and C.'
    FLUSH(output_FID)
    !
    basis_vector_a = vector1
    basis_vector_b = vector2
    basis_vector_c = vector3
 ELSE IF (num_periodic_dir .EQ. 2) THEN  
    max_coords = -HUGE(1.0_dp)
    min_coords = HUGE(1.0_dp)
    IF (.NOT. periodicA) THEN
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along directions B and C. Not periodic along direction A.'
       FLUSH(output_FID)
       !
       basis_vector_a = cross(vector2,vector3)
       normalization = SQRT(DOT_PRODUCT(basis_vector_a,basis_vector_a))
       pre_basis_vector_a = basis_vector_a/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_a,xyz)
          max_coords(1) = max(temp1,max_coords(1))
          min_coords(1) = min(temp1,min_coords(1))
       END DO
       basis_vector_a = pre_basis_vector_a*(1.0_dp + max_coords(1) - min_coords(1)) 
       basis_vector_b = vector2
       basis_vector_c = vector3      
    ELSE IF (.NOT. periodicB) THEN
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along directions A and C. Not periodic along direction B.'
       FLUSH(output_FID)
       !
       basis_vector_a = vector1
       basis_vector_b = cross(vector3,vector1)
       normalization = SQRT(DOT_PRODUCT(basis_vector_b,basis_vector_b))
       pre_basis_vector_b = basis_vector_b/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_b,xyz)
          max_coords(2) = max(temp1,max_coords(2))
          min_coords(2) = min(temp1,min_coords(2))
       END DO
       basis_vector_b = pre_basis_vector_b*(1.0_dp + max_coords(2) - min_coords(2))
       basis_vector_c = vector3 
     ELSE
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along directions A and B. Not periodic along direction C.'
       FLUSH(output_FID)
       !
       basis_vector_a = vector1
       basis_vector_b = vector2
       basis_vector_c = cross(vector1,vector2)
       normalization = SQRT(DOT_PRODUCT(basis_vector_c,basis_vector_c))
       pre_basis_vector_c = basis_vector_c/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_c,xyz)
          max_coords(3) = max(temp1,max_coords(3))
          min_coords(3) = min(temp1,min_coords(3))
       END DO
       basis_vector_c = pre_basis_vector_c*(1.0_dp + max_coords(3) - min_coords(3))
    END IF    
 ELSE IF (num_periodic_dir .EQ. 1) THEN  
    max_coords = -HUGE(1.0_dp)
    min_coords = HUGE(1.0_dp)
    IF (periodicA) THEN
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along direction A. Not periodic along directions B and C.'
       FLUSH(output_FID)
       !
       basis_vector_a = vector1
       IF (abs(basis_vector_a(1)) > abs(basis_vector_a(3))) THEN
          basis_vector_b=cross([0.0_dp, 0.0_dp, 1.0_dp],basis_vector_a)
       ELSE
          basis_vector_b=cross(basis_vector_a,[1.0_dp, 0.0_dp, 0.0_dp])
       END IF
       normalization = SQRT(DOT_PRODUCT(basis_vector_b,basis_vector_b))
       pre_basis_vector_b = basis_vector_b/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_b,xyz)
          max_coords(2) = max(temp1,max_coords(2))
          min_coords(2) = min(temp1,min_coords(2))
       END DO
       basis_vector_b = pre_basis_vector_b*(1.0_dp + max_coords(2) - min_coords(2)) 
       basis_vector_c = cross(basis_vector_a,basis_vector_b)
       normalization = SQRT(DOT_PRODUCT(basis_vector_c,basis_vector_c))
       pre_basis_vector_c = basis_vector_c/normalization 
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_c,xyz)
          max_coords(3) = max(temp1,max_coords(3))
          min_coords(3) = min(temp1,min_coords(3))
       END DO
       basis_vector_c = pre_basis_vector_c*(1.0_dp + max_coords(3) - min_coords(3))        
    ELSE IF (periodicB) THEN
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along direction B. Not periodic along directions A and C.'
       FLUSH(output_FID)
       !
       basis_vector_b = vector2
       IF (abs(basis_vector_b(1)) > abs(basis_vector_b(3))) THEN
          basis_vector_a=cross([0.0_dp, 0.0_dp, 1.0_dp],basis_vector_b)
       ELSE
          basis_vector_a=cross(basis_vector_b,[1.0_dp, 0.0_dp, 0.0_dp])
       END IF
       normalization = SQRT(DOT_PRODUCT(basis_vector_a,basis_vector_a))
       pre_basis_vector_a = basis_vector_a/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_a,xyz)
          max_coords(1) = max(temp1,max_coords(1))
          min_coords(1) = min(temp1,min_coords(1))
       END DO
       basis_vector_a = pre_basis_vector_a*(1.0_dp + max_coords(1) - min_coords(1))
       basis_vector_c = cross(basis_vector_a,basis_vector_b)
       normalization = SQRT(DOT_PRODUCT(basis_vector_c,basis_vector_c))
       pre_basis_vector_c = basis_vector_c/normalization  
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_c,xyz)
          max_coords(3) = max(temp1,max_coords(3))
          min_coords(3) = min(temp1,min_coords(3))
       END DO
       basis_vector_c = pre_basis_vector_c*(1.0_dp + max_coords(3) - min_coords(3))
    ELSE
       !
       WRITE(output_FID,*) ' '
       WRITE(output_FID,*) 'Periodic along direction C. Not periodic along directions A and B.'
       FLUSH(output_FID)
       !
       basis_vector_c = vector3 
       IF (abs(basis_vector_c(1)) > abs(basis_vector_c(3))) THEN
          basis_vector_a=cross([0.0_dp, 0.0_dp, 1.0_dp],basis_vector_c)
       ELSE
          basis_vector_a=cross(basis_vector_c,[1.0_dp, 0.0_dp, 0.0_dp])
       END IF
       normalization = SQRT(DOT_PRODUCT(basis_vector_a,basis_vector_a))
       pre_basis_vector_a = basis_vector_a/normalization
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_a,xyz)
          max_coords(1) = max(temp1,max_coords(1))
          min_coords(1) = min(temp1,min_coords(1))
       END DO
       basis_vector_a = pre_basis_vector_a*(1.0_dp + max_coords(1) - min_coords(1)) 
       basis_vector_b = cross(basis_vector_c,basis_vector_a)
       normalization = SQRT(DOT_PRODUCT(basis_vector_b,basis_vector_b))
       pre_basis_vector_b = basis_vector_b/normalization  
       DO i=1,natoms
          xyz(:) = coords(:,i)
          temp1 = DOT_PRODUCT(basis_vector_b,xyz)
          max_coords(2) = max(temp1,max_coords(2))
          min_coords(2) = min(temp1,min_coords(2))
       END DO
       basis_vector_b = pre_basis_vector_b*(1.0_dp + max_coords(2) - min_coords(2))
    END IF 
 ELSE
    !
    WRITE(output_FID,*) ' '
    WRITE(output_FID,*) 'Not periodic along directions A, B, and C.'
    FLUSH(output_FID)
    !
    max_coords = MAXVAL(coords,2)
    min_coords = MINVAL(coords,2)
    pre_basis_vector_a(1) =  1.0_dp
    pre_basis_vector_b(2) =  1.0_dp 
    pre_basis_vector_c(3) =  1.0_dp
    basis_vector_a(1) =  max_coords(1) - min_coords(1) + 1.0_dp
    basis_vector_b(2) =  max_coords(2) - min_coords(2) + 1.0_dp 
    basis_vector_c(3) =  max_coords(3) - min_coords(3) + 1.0_dp
 END IF

 ! Compute the unit cell volume
 basis_vectors(1,1) = basis_vector_a(1)
 basis_vectors(1,2) = basis_vector_a(2)
 basis_vectors(1,3) = basis_vector_a(3)
 basis_vectors(2,1) = basis_vector_b(1)
 basis_vectors(2,2) = basis_vector_b(2)
 basis_vectors(2,3) = basis_vector_b(3)
 basis_vectors(3,1) = basis_vector_c(1)
 basis_vectors(3,2) = basis_vector_c(2)
 basis_vectors(3,3) = basis_vector_c(3)
 unit_cell_volume = abs(determinant(basis_vectors,3))
 
 WRITE(output_FID,*)'The parallelepiped enclosing the reference unit cell or molecule in atomic units is'
 DO i=1,3
   WRITE(output_FID, *) (basis_vectors(i,j),j=1,3)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 inverse_vectors=matrixinv(TRANSPOSE(basis_vectors),3)

 WRITE(output_FID,*)'The reciprocal space vectors in atomic units are'
 DO i=1,3
   WRITE(output_FID, *) (inverse_vectors(i,j),j=1,3)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)
 
 ! Define the origin
 DO i = 1,natoms
   xyz(:)=coords(:,i) 
   inverse_coords(:,i) = matmul(inverse_vectors,xyz)
   IF (i .EQ. 1) THEN
      IF (periodicA) THEN
         min_coords(1) = 0.0_dp
      ELSE
         min_coords(1) = inverse_coords(1,1)
      END IF
      IF (periodicB) THEN
         min_coords(2) = 0.0_dp
      ELSE
         min_coords(2) = inverse_coords(2,1)
      END IF
      IF (periodicC) THEN
         min_coords(3) = 0.0_dp
      ELSE
         min_coords(3) = inverse_coords(3,1)
      END IF
   END IF
   IF (.NOT. periodicA) min_coords(1) = min(min_coords(1),inverse_coords(1,i))  
   IF (.NOT. periodicB) min_coords(2) = min(min_coords(2),inverse_coords(2,i))    
   IF (.NOT. periodicC) min_coords(3) = min(min_coords(3),inverse_coords(3,i))      
 END DO
 DO j= 1,3
    origin(j) = min_coords(1)*basis_vector_a(j) + min_coords(2)*basis_vector_b(j) + min_coords(3)*basis_vector_c(j) &
    - 0.5_dp*pre_basis_vector_a(j) - 0.5_dp*pre_basis_vector_b(j) - 0.5_dp*pre_basis_vector_c(j)
 END DO 

 WRITE(output_FID,*)'Origin ='
 WRITE(output_FID, *) (origin(j),j=1,3)
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 ! Recompute the coordinates and reciprocal space coordinates inside the reference unit cell
 DO i = 1,natoms
   xyz(:)=coords(:,i)
   inverse_coords(:,i) = matmul(inverse_vectors,(xyz - origin))
   inverse_coords(:,i) = min(max(modulo(inverse_coords(:,i),1.0_dp),0.0_dp + zero_tolerance),1.0_dp - zero_tolerance)
   DO j= 1,3
      coords(j,i) = inverse_coords(1,i)*basis_vector_a(j) + inverse_coords(2,i)*basis_vector_b(j)&
      + inverse_coords(3,i)*basis_vector_c(j) + origin(j)
   END DO
 END DO

 WRITE(output_FID,*)'Adjusted coords (for up to the first 100 atoms) in atomic units ='
 DO i=1,min(natoms,100)
   WRITE(output_FID, *) (coords(j,i),j=1,3)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 WRITE(output_FID,*)'inverse_coords (for up to the first 100 atoms) in atomic units ='
 DO i=1,min(natoms,100)
   WRITE(output_FID, *) (inverse_coords(j,i),j=1,3)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)
 
 ! ==========================================================================
 ! STEP # 2: Divide the unit cell parallelepiped into spatial regions
 ! ==========================================================================

 ! Compute the basis vector dot products and number of regions along each basis vector
 length_a = sqrt(basis_vector_a(1)**2 + basis_vector_a(2)**2 + basis_vector_a(3)**2)
 length_b = sqrt(basis_vector_b(1)**2 + basis_vector_b(2)**2 + basis_vector_b(3)**2)
 length_c = sqrt(basis_vector_c(1)**2 + basis_vector_c(2)**2 + basis_vector_c(3)**2)
 a_dot_b = basis_vector_a(1)*basis_vector_b(1) + basis_vector_a(2)*basis_vector_b(2) + basis_vector_a(3)*basis_vector_b(3)
 a_dot_c = basis_vector_a(1)*basis_vector_c(1) + basis_vector_a(2)*basis_vector_c(2) + basis_vector_a(3)*basis_vector_c(3)
 b_dot_c = basis_vector_b(1)*basis_vector_c(1) + basis_vector_b(2)*basis_vector_c(2) + basis_vector_b(3)*basis_vector_c(3)
 cos_a_b = (a_dot_b)/(length_a*length_b)
 sin_a_b = sqrt(1.0_dp - cos_a_b*cos_a_b)
 cos_a_c = (a_dot_c)/(length_a*length_c)
 sin_a_c = sqrt(1.0_dp - cos_a_c*cos_a_c) 
 cos_b_c = (b_dot_c)/(length_b*length_c)
 sin_b_c = sqrt(1.0_dp - cos_b_c*cos_b_c)
 ref_regions_A = ceiling(3.0_dp*unit_cell_volume/(dipole_interaction_cutoff_length*length_b*length_c*sin_b_c))
 ref_regions_B = ceiling(3.0_dp*unit_cell_volume/(dipole_interaction_cutoff_length*length_a*length_c*sin_a_c))
 ref_regions_C = ceiling(3.0_dp*unit_cell_volume/(dipole_interaction_cutoff_length*length_a*length_b*sin_a_b))

 WRITE(output_FID,*)'The unit cell has been divided into the following number of regions along each real-space direction:'
 WRITE(output_FID, *) ref_regions_A,ref_regions_B,ref_regions_C
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 WRITE(output_FID,*)'Sorting atoms in the reference unit cell according to spatial regions ...'
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)
 
 ! Classify and sort atoms of the reference unit cell into regions
 ALLOCATE(ref_region_count(ref_regions_A,ref_regions_B,ref_regions_C))
 ref_region_count = 0
 DO atom1=1,natoms
    region_A = NINT(ref_regions_A*inverse_coords(1,atom1) + 0.5_dp)
    region_B = NINT(ref_regions_B*inverse_coords(2,atom1) + 0.5_dp)
    region_C = NINT(ref_regions_C*inverse_coords(3,atom1) + 0.5_dp)
    ref_region_count(region_A,region_B,region_C) = ref_region_count(region_A,region_B,region_C) + 1
 END DO

 ALLOCATE(region_start_end(5,ref_regions_A,ref_regions_B,ref_regions_C))
 ALLOCATE(temp_region_count(ref_regions_A,ref_regions_B,ref_regions_C))
 region_start_end = 0
 temp_count = 1
 DO region_C = 1,ref_regions_C
    DO region_B = 1,ref_regions_B
       DO region_A = 1,ref_regions_A
          region_start_end(1,region_A,region_B,region_C) =  temp_count
          temp_region_count(region_A,region_B,region_C) =  temp_count
          temp_count = temp_count + ref_region_count(region_A,region_B,region_C)
          region_start_end(2,region_A,region_B,region_C) = temp_count - 1
       END DO
    END DO
 END DO

 DO atom1=1,natoms
    region_A = NINT(ref_regions_A*inverse_coords(1,atom1) + 0.5_dp)
    region_B = NINT(ref_regions_B*inverse_coords(2,atom1) + 0.5_dp)
    region_C = NINT(ref_regions_C*inverse_coords(3,atom1) + 0.5_dp)
    temp_count = temp_region_count(region_A,region_B,region_C)
    sorted_atoms(1,temp_count) = atom1
    sorted_atoms(2,temp_count) = region_A   
    sorted_atoms(3,temp_count) = region_B
    sorted_atoms(4,temp_count) = region_C  
    sorted_coords(1,temp_count) = atom1 
    sorted_coords(2,temp_count) = coords(1,atom1)
    sorted_coords(3,temp_count) = coords(2,atom1)
    sorted_coords(4,temp_count) = coords(3,atom1)
    temp_region_count(region_A,region_B,region_C) = &
    temp_region_count(region_A,region_B,region_C)  +  1
 END DO

 DEALLOCATE(inverse_coords)
 DEALLOCATE(temp_region_count)
 DEALLOCATE(ref_region_count)

 WRITE(output_FID,*)'sorted_coords (for up to the first 100 atoms) in atomic units ='
 WRITE(output_FID,*)'atom number, X, Y, Z, atom number, region A, region B, region C'
 FLUSH(output_FID)
 DO i=1,min(natoms,100)
    WRITE(output_FID, *) (sorted_coords(j,i),j=1,4),(sorted_atoms(j,i),j=1,4)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)
 IF (natoms > 199) THEN
    WRITE(output_FID,*)'sorted_coords (for the last 100 atoms) in atomic units ='
    WRITE(output_FID,*)'atom number, X, Y, Z, atom number, region A, region B, region C'
    DO i=1,100
       WRITE(output_FID, *) (sorted_coords(j,(natoms - 100 + i)),j=1,4),(sorted_atoms(j,(natoms - 100 + i)),j=1,4)
    END DO
    WRITE(output_FID, *) ' '
    FLUSH(output_FID)
 END IF
 
 ! ==========================================================================
 ! STEP # 3: Construct arrays listing interacting region pair images
 ! ==========================================================================

 ! Compute the periodic sum limits along each direction
 IF (periodicA) THEN
   periodicsumlimitA = ceiling(dipole_interaction_cutoff_length*length_b*length_c*sin_b_c/unit_cell_volume)
 ELSE
   periodicsumlimitA = 0
 END IF
 IF (periodicB) THEN
   periodicsumlimitB = ceiling(dipole_interaction_cutoff_length*length_a*length_c*sin_a_c/unit_cell_volume)
 ELSE
   periodicsumlimitB = 0
 END IF
 IF (periodicC) THEN
   periodicsumlimitC = ceiling(dipole_interaction_cutoff_length*length_a*length_b*sin_a_b/unit_cell_volume)
 ELSE
   periodicsumlimitC = 0
 END IF

 IF (ignore_PBC) THEN
  periodicsumlimitA = 0
  periodicsumlimitB = 0
  periodicsumlimitC = 0
 END IF

 WRITE(Output_FID,*)'The computed periodic sum limits along each lattice direction are: ',periodicsumlimitA,&
 periodicsumlimitB,periodicsumlimitC
 WRITE(Output_FID,*)' '
 FLUSH(Output_FID)

 ! Compute the longest body diagonal of one spatial region
 longest_body_diag = 0.0_dp
 delta_a = 1.0_dp/ref_regions_A
 delta_b = 1.0_dp/ref_regions_B
 delta_c = 1.0_dp/ref_regions_C
 X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
 Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
 Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
 distance = sqrt(X1**2 + Y1**2 + Z1**2)
 longest_body_diag = max(longest_body_diag,distance)
 delta_a = -1.0_dp/ref_regions_A
 delta_b = 1.0_dp/ref_regions_B
 delta_c = 1.0_dp/ref_regions_C
 X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
 Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
 Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
 distance = sqrt(X1**2 + Y1**2 + Z1**2)
 longest_body_diag = max(longest_body_diag,distance)
 delta_a = 1.0_dp/ref_regions_A
 delta_b = -1.0_dp/ref_regions_B
 delta_c = 1.0_dp/ref_regions_C
 X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
 Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
 Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
 distance = sqrt(X1**2 + Y1**2 + Z1**2)
 longest_body_diag = max(longest_body_diag,distance)
 delta_a = 1.0_dp/ref_regions_A
 delta_b = 1.0_dp/ref_regions_B
 delta_c = -1.0_dp/ref_regions_C
 X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
 Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
 Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
 distance = sqrt(X1**2 + Y1**2 + Z1**2)
 longest_body_diag = max(longest_body_diag,distance)

 WRITE(output_FID,*)'Counting the numbers of interacting spatial regions and their periodic images ...'
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 ! Prepare the list of interacting regions and their periodic images
 temp_count = 0
 num_region_pairs = 0
 dist_squared_cutoff = (dipole_interaction_cutoff_length + longest_body_diag)**2
 DO region_C = 1,ref_regions_C
    IF (ref_regions_C .LE. 11) THEN
       start_k = 1
       end_k = ref_regions_C
    ELSE
       start_k = region_C - 5
       end_k = region_C + 5
    END IF
    DO region_B = 1,ref_regions_B
       IF (ref_regions_B .LE. 11) THEN
          start_j = 1
          end_j = ref_regions_B
       ELSE
          start_j = region_B - 5
          end_j = region_B + 5
       END IF
       DO region_A = 1,ref_regions_A
          ! Skip if the region contains no atoms
          IF (region_start_end(2,region_A,region_B,region_C) < region_start_end(1,region_A,region_B,region_C)) CYCLE
          prior_num_region_pairs = num_region_pairs
          IF (ref_regions_A .LE. 11) THEN
             start_i = 1
             end_i = ref_regions_A
          ELSE
             start_i = region_A - 5
             end_i = region_A + 5
          END IF
          DO index_k = start_k,end_k
             IF ( (.NOT. periodicC) .AND. ( (index_k < 1) .OR. (index_k > ref_regions_C) ) ) CYCLE
             k = modulo(index_k - 1,ref_regions_C) + 1
             DO index_j = start_j,end_j
                IF ( (.NOT. periodicB) .AND. ( (index_j < 1) .OR. (index_j > ref_regions_B) ) ) CYCLE
                j = modulo(index_j - 1,ref_regions_B) + 1
                DO index_i = start_i,end_i
                   IF ( (.NOT. periodicA) .AND. ( (index_i < 1) .OR. (index_i > ref_regions_A) ) ) CYCLE
                   i = modulo(index_i - 1,ref_regions_A) + 1
                   ! Skip if the region contains no atoms
                   IF (region_start_end(2,i,j,k) < region_start_end(1,i,j,k)) CYCLE
                   active_region = .false.
                   DO repeata = -periodicsumlimitA,periodicsumlimitA
                      delta_a = (i - region_A)/REAL(ref_regions_A,dp) + repeata
                      DO repeatb = -periodicsumlimitB,periodicsumlimitB
                         delta_b = (j - region_B)/REAL(ref_regions_B,dp) + repeatb
                         DO repeatc = -periodicsumlimitC,periodicsumlimitC
                            delta_c = (k - region_C)/REAL(ref_regions_C,dp) + repeatc
                            X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
                            Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
                            Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
                            dist_squared = X1**2 + Y1**2 + Z1**2
                            IF (dist_squared <= dist_squared_cutoff) THEN
                               active_region = .true. 
                               temp_count =  temp_count + 1
                            END IF
                         END DO
                      END DO
                   END DO
                   IF (active_region) num_region_pairs = num_region_pairs + 1
                END DO
             END DO 
          END DO 
          IF (prior_num_region_pairs < num_region_pairs) THEN
             region_start_end(3,region_A,region_B,region_C) = prior_num_region_pairs + 1
             region_start_end(4,region_A,region_B,region_C) = num_region_pairs
             region_start_end(5,region_A,region_B,region_C) = 1
          END IF
       END DO
    END DO
 END DO

 WRITE(output_FID,*)'region_start_end (for up to three regions along each lattice direction) = '
 WRITE(output_FID,*)'column 1 = sorted atom start index'
 WRITE(output_FID,*)'column 2 = sorted atom end index'
 WRITE(output_FID,*)'column 3 = region pair start index'
 WRITE(output_FID,*)'column 4 = region pair end index'
 WRITE(output_FID,*)'column 5 = whether region has any active pairs (1) or not (0)'
 FLUSH(output_FID)  
 DO region_C = 1,min(ref_regions_C,3)
    DO region_B = 1,min(ref_regions_B,3)
       DO region_A = 1,min(ref_regions_A,3)
          WRITE(output_FID, *) 'region A = ',region_A,'region B = ',region_B,'region C = ',region_C,&
          'region_start_end = ',(region_start_end(j,region_A,region_B,region_C),j=1,5)
       END DO
    END DO
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)  

 WRITE(output_FID, *) ' '
 WRITE(output_FID,*)'Preparing the lists of interacting spatial regions and their periodic images ...'
 WRITE(output_FID,*)'num_region_pairs =',num_region_pairs
 WRITE(output_FID,*)'Total number of interacting image region pairs =',temp_count
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 ALLOCATE(region_pairs(8,num_region_pairs))
 ALLOCATE(region_pair_images(4,temp_count))

 temp_count = 0
 num_region_pairs = 0

 DO region_C = 1,ref_regions_C
    IF (ref_regions_C .LE. 11) THEN
       start_k = 1
       end_k = ref_regions_C
    ELSE
       start_k = region_C - 5
       end_k = region_C + 5
    END IF
    DO region_B = 1,ref_regions_B
       IF (ref_regions_B .LE. 11) THEN
          start_j = 1
          end_j = ref_regions_B
       ELSE
          start_j = region_B - 5
          end_j = region_B + 5
       END IF
       DO region_A = 1,ref_regions_A
          ! Skip if the region contains no atoms
          IF (region_start_end(2,region_A,region_B,region_C) < region_start_end(1,region_A,region_B,region_C)) CYCLE
          prior_num_region_pairs = num_region_pairs
          IF (ref_regions_A .LE. 11) THEN
             start_i = 1
             end_i = ref_regions_A
          ELSE
             start_i = region_A - 5
             end_i = region_A + 5
          END IF
          DO index_k = start_k,end_k
             IF ( (.NOT. periodicC) .AND. ( (index_k < 1) .OR. (index_k > ref_regions_C) ) ) CYCLE
             k = modulo(index_k - 1,ref_regions_C) + 1
             DO index_j = start_j,end_j
                IF ( (.NOT. periodicB) .AND. ( (index_j < 1) .OR. (index_j > ref_regions_B) ) ) CYCLE
                j = modulo(index_j - 1,ref_regions_B) + 1
                DO index_i = start_i,end_i
                   IF ( (.NOT. periodicA) .AND. ( (index_i < 1) .OR. (index_i > ref_regions_A) ) ) CYCLE
                   i = modulo(index_i - 1,ref_regions_A) + 1
                   ! Skip if the region contains no atoms
                   IF (region_start_end(2,i,j,k) < region_start_end(1,i,j,k)) CYCLE
                   prior_temp_count = temp_count
                   active_region = .false.
                   DO repeata = -periodicsumlimitA,periodicsumlimitA
                      delta_a = (i - region_A)/REAL(ref_regions_A,dp) + repeata
                      DO repeatb = -periodicsumlimitB,periodicsumlimitB
                         delta_b = (j - region_B)/REAL(ref_regions_B,dp) + repeatb
                         DO repeatc = -periodicsumlimitC,periodicsumlimitC
                            delta_c = (k - region_C)/REAL(ref_regions_C,dp) + repeatc
                            X1 = delta_a*basis_vector_a(1) + delta_b*basis_vector_b(1) + delta_c*basis_vector_c(1)
                            Y1 = delta_a*basis_vector_a(2) + delta_b*basis_vector_b(2) + delta_c*basis_vector_c(2)
                            Z1 = delta_a*basis_vector_a(3) + delta_b*basis_vector_b(3) + delta_c*basis_vector_c(3)
                            dist_squared = X1**2 + Y1**2 + Z1**2
                            IF (dist_squared <= dist_squared_cutoff) THEN
                               active_region = .true. 
                               temp_count =  temp_count + 1
                               region_pair_images(1,temp_count) = repeata
                               region_pair_images(2,temp_count) = repeatb
                               region_pair_images(3,temp_count) = repeatc
                               region_pair_images(4,temp_count) = num_region_pairs + 1
                            END IF
                         END DO
                      END DO
                   END DO
                   IF (active_region) THEN
                      num_region_pairs = num_region_pairs + 1
                      region_pairs(1,num_region_pairs) = region_A
                      region_pairs(2,num_region_pairs) = region_B
                      region_pairs(3,num_region_pairs) = region_C
                      region_pairs(4,num_region_pairs) = i
                      region_pairs(5,num_region_pairs) = j
                      region_pairs(6,num_region_pairs) = k 
                      region_pairs(7,num_region_pairs) = prior_temp_count + 1
                      region_pairs(8,num_region_pairs) = temp_count                   
                   END IF
                END DO
             END DO 
          END DO 
       END DO
    END DO
 END DO

 WRITE(output_FID,*)'region_pairs entry (for up to the first 100 rows) ='
 WRITE(output_FID,*)'region A, region B, region C, region i, region j, region k, start image index, end image index'
 DO i=1,min(num_region_pairs,100)
    WRITE(output_FID, *) (region_pairs(j,i),j=1,8)
 END DO
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)  

 ! ==========================================================================
 ! STEP # 4: Construct two lists of interacting atom pairs
 ! ==========================================================================

 WRITE(output_FID,*)'Counting the numbers of interacting atom pairs ...'
 WRITE(output_FID, *) ' '
 FLUSH(output_FID)

 !Determine the number of symmetry unique atom pairs to include in the atom image pair matrices
 temp_count_large = 0
 temp_count_small = 0
 pair_count_small = 0
 pair_count_large = 0

!$omp parallel do default(none) &
!$omp private(index_1,atom1,region_A,region_B,region_C,X1,Y1,Z1,index_2,i,j,k,index_3,atom2,temp_X2,temp_Y2,temp_Z2, &
!$omp attenuation_length,active_pair_large,index_4,repeata,repeatb,repeatc,X2,Y2,Z2,distance,distance_to_attenuation_ratio) &
!$omp shared(natoms,sorted_atoms,region_start_end,sorted_coords,region_pairs,static_unscreened_atomic_polarizability,wp, &
!$omp region_pair_images,vector1,vector2,vector3,dipole_interaction_cutoff_length,distance_to_attenuation_ratio_cutoff, &
!$omp pair_count_small,pair_count_large) &
!$omp schedule(static)
 DO index_1 = 1,natoms
    atom1 = sorted_atoms(1,index_1) 
    region_A = sorted_atoms(2,index_1) 
    region_B = sorted_atoms(3,index_1) 
    region_C = sorted_atoms(4,index_1) 
    IF (region_start_end(5,region_A,region_B,region_C) == 0) CYCLE
    X1=sorted_coords(2,index_1)
    Y1=sorted_coords(3,index_1)
    Z1=sorted_coords(4,index_1)
    DO index_2 = region_start_end(3,region_A,region_B,region_C),region_start_end(4,region_A,region_B,region_C)
       i = region_pairs(4,index_2)
       j = region_pairs(5,index_2) 
       k = region_pairs(6,index_2)
       DO index_3 = region_start_end(1,i,j,k),region_start_end(2,i,j,k)
          atom2 = sorted_atoms(1,index_3)
          IF (atom2 < atom1) CYCLE  
          temp_X2=sorted_coords(2,index_3)
          temp_Y2=sorted_coords(3,index_3)
          temp_Z2=sorted_coords(4,index_3)
          attenuation_length = calculate_attenuation_length(0.0_dp,static_unscreened_atomic_polarizability(atom1),&
          static_unscreened_atomic_polarizability(atom2),wp(atom1),wp(atom2))
          active_pair_large = .false.    
          DO index_4 = region_pairs(7,index_2),region_pairs(8,index_2)  
             repeata = region_pair_images(1,index_4) 
             repeatb = region_pair_images(2,index_4) 
             repeatc = region_pair_images(3,index_4)   
             X2=temp_X2 + repeata*vector1(1) + repeatb*vector2(1) + repeatc*vector3(1)
             Y2=temp_Y2 + repeata*vector1(2) + repeatb*vector2(2) + repeatc*vector3(2)
             Z2=temp_Z2 + repeata*vector1(3) + repeatb*vector2(3) + repeatc*vector3(3)
             distance = sqrt((X1-X2)**2 + (Y1-Y2)**2 + (Z1-Z2)**2)
             IF (distance > dipole_interaction_cutoff_length) THEN
                CYCLE
             !determine if the pair is symmetrically unique
             ELSE IF (atom2 > atom1) THEN
                active_pair_large = .true.
             ELSE IF (repeata > 0) THEN
                active_pair_large = .true.
             ELSE IF ((repeata == 0) .and. (repeatb > 0)) THEN
                active_pair_large = .true.
             ELSE IF (((repeata == 0) .and. (repeatb == 0)) .and. (repeatc > 0)) THEN
                active_pair_large = .true.
             ELSE
                CYCLE
             END IF
             distance_to_attenuation_ratio = distance/attenuation_length
             IF (distance_to_attenuation_ratio <= distance_to_attenuation_ratio_cutoff) THEN
                pair_count_small(atom1) = pair_count_small(atom1) + 1
             END IF    
          END DO
          IF (active_pair_large) THEN
             pair_count_large(atom1) = pair_count_large(atom1) + 1
          END IF
       END DO
    END DO
 END DO
!$omp end parallel do

 WRITE(output_FID,*)'The total number of symmetry unique atom image pairs to include in the atom image pair matrix is'
 num_sym_unique_atom_image_pairs_large = sum(pair_count_large)
 num_sym_unique_atom_image_pairs_small = sum(pair_count_small)
 WRITE(output_FID,*)'num_sym_unique_atom_image_pairs_large: ',num_sym_unique_atom_image_pairs_large
 WRITE(output_FID,*)'num_sym_unique_atom_image_pairs_small: ',num_sym_unique_atom_image_pairs_small
 FLUSH(output_FID)
 IF (num_sym_unique_atom_image_pairs_small == 0) THEN
   WRITE(output_FID,*)'The are no symmetry unique atom-image pairs to include in the atom image pair matrix.'
   WRITE(output_FID,*)'Because there are no images acting on the unit cell, the screened analysis will be skipped.'
   RETURN
 END IF

 !Initialize the atom image pair matrices
 ALLOCATE(atom_image_pair_matrix_large(16,num_sym_unique_atom_image_pairs_large))
 ALLOCATE(atom_image_pair_matrix_small(16,num_sym_unique_atom_image_pairs_small))
 temp_count_small = pair_count_small(1)
 temp_count_large = pair_count_large(1)
 pair_count_small(1) = 0
 pair_count_large(1) = 0
 !Compute the cummulative totals for atoms 1 to (i-1) and store in pair_count_small(i) and pair_count_large(i)
 DO i=2,natoms
    j=pair_count_small(i)
    k=pair_count_large(i)
    pair_count_small(i) = pair_count_small(i-1) + temp_count_small
    pair_count_large(i) = pair_count_large(i-1) + temp_count_large
    temp_count_small = j
    temp_count_large = k
 END DO

 atom_image_pair_matrix_large=0.0_dp
 atom_image_pair_matrix_small=0.0_dp

!$omp parallel do default(none) &
!$omp private(index_1,atom1,region_A,region_B,region_C,X1,Y1,Z1,index_2,i,j,k,index_3,atom2,temp_X2,temp_Y2,temp_Z2, &
!$omp attenuation_length,active_pair_large,index_4,repeata,repeatb,repeatc,X2,Y2,Z2,distance,distance_to_attenuation_ratio, &
!$omp temp_count_small,temp_count_large,active_pair_small,total_radius,temp1,temp2,temp4,temp5,temp6,delta_x,delta_y,delta_z) &
!$omp shared(natoms,sorted_atoms,region_start_end,sorted_coords,region_pairs,static_unscreened_atomic_polarizability,wp, &
!$omp region_pair_images,vector1,vector2,vector3,dipole_interaction_cutoff_length,distance_to_attenuation_ratio_cutoff, &
!$omp pair_count_small,pair_count_large,unscreened_damping_radius,multibody_screening_parameter, &
!$omp atom_image_pair_matrix_large,atom_image_pair_matrix_small) &
!$omp schedule(static)
 DO index_1 = 1,natoms
    atom1 = sorted_atoms(1,index_1) 
    region_A = sorted_atoms(2,index_1) 
    region_B = sorted_atoms(3,index_1) 
    region_C = sorted_atoms(4,index_1) 
    IF (region_start_end(5,region_A,region_B,region_C) == 0) CYCLE
    X1=sorted_coords(2,index_1)
    Y1=sorted_coords(3,index_1)
    Z1=sorted_coords(4,index_1)
    temp_count_small = pair_count_small(atom1) 
    temp_count_large = pair_count_large(atom1) + 1
    DO index_2 = region_start_end(3,region_A,region_B,region_C),region_start_end(4,region_A,region_B,region_C)
       i = region_pairs(4,index_2)
       j = region_pairs(5,index_2) 
       k = region_pairs(6,index_2) 
       DO index_3 = region_start_end(1,i,j,k),region_start_end(2,i,j,k)
          atom2 = sorted_atoms(1,index_3)
          IF (atom2 < atom1) CYCLE     
          temp_X2=sorted_coords(2,index_3)
          temp_Y2=sorted_coords(3,index_3)
          temp_Z2=sorted_coords(4,index_3)
          attenuation_length = calculate_attenuation_length(0.0_dp,static_unscreened_atomic_polarizability(atom1),&
          static_unscreened_atomic_polarizability(atom2),wp(atom1),wp(atom2))
          active_pair_large = .false. 
          DO index_4 = region_pairs(7,index_2),region_pairs(8,index_2)  
             repeata = region_pair_images(1,index_4) 
             repeatb = region_pair_images(2,index_4) 
             repeatc = region_pair_images(3,index_4)
             active_pair_small = .false.   
             X2=temp_X2 + repeata*vector1(1) + repeatb*vector2(1) + repeatc*vector3(1)
             Y2=temp_Y2 + repeata*vector1(2) + repeatb*vector2(2) + repeatc*vector3(2)
             Z2=temp_Z2 + repeata*vector1(3) + repeatb*vector2(3) + repeatc*vector3(3)
             distance = sqrt((X1-X2)**2 + (Y1-Y2)**2 + (Z1-Z2)**2)
             IF (distance > dipole_interaction_cutoff_length) THEN
                CYCLE
             !determine if the pair is symmetrically unique
             ELSE IF (atom2 > atom1) THEN
                active_pair_large = .true.
             ELSE IF (repeata > 0) THEN
                active_pair_large = .true.
             ELSE IF ((repeata == 0) .and. (repeatb > 0)) THEN
                active_pair_large = .true.
             ELSE IF (((repeata == 0) .and. (repeatb == 0)) .and. (repeatc > 0)) THEN
                active_pair_large = .true.
             ELSE
                CYCLE
             END IF
             distance_to_attenuation_ratio = distance/attenuation_length
             total_radius = bohrperangstrom*(unscreened_damping_radius(atom1) + unscreened_damping_radius(atom2))
             IF (distance_to_attenuation_ratio <= distance_to_attenuation_ratio_cutoff) THEN
                temp_count_small = temp_count_small + 1
                active_pair_small = .true.
             END IF  
             !initialize the atom pair matrix
             temp1 = smooth_cutoff_function(distance,dipole_interaction_cutoff_length)
             temp2 = temp1/distance**5
             delta_x = X2-X1
             delta_y = Y2-Y1
             delta_z = Z2-Z1
             atom_image_pair_matrix_large(1,temp_count_large) = atom1
             atom_image_pair_matrix_large(2,temp_count_large) = atom2
             ! Accumulations for the static polarizability
             atom_image_pair_matrix_large(3,temp_count_large) = atom_image_pair_matrix_large(3,temp_count_large) &
             + temp2*(distance**2 - 3.0_dp*delta_x**2)
             atom_image_pair_matrix_large(4,temp_count_large) = atom_image_pair_matrix_large(4,temp_count_large) &
             - 3.0_dp*temp2*delta_x*delta_y
             atom_image_pair_matrix_large(5,temp_count_large) = atom_image_pair_matrix_large(5,temp_count_large) &
             - 3.0_dp*temp2*delta_x*delta_z
             atom_image_pair_matrix_large(6,temp_count_large) = atom_image_pair_matrix_large(6,temp_count_large) &
             + temp2*(distance**2 - 3.0_dp*delta_y**2)
             atom_image_pair_matrix_large(7,temp_count_large) = atom_image_pair_matrix_large(7,temp_count_large) &
             - 3.0_dp*temp2*delta_y*delta_z
             atom_image_pair_matrix_large(8,temp_count_large) = atom_image_pair_matrix_large(8,temp_count_large) &
             + temp2*(distance**2 - 3.0_dp*delta_z**2)
             atom_image_pair_matrix_large(9,temp_count_large) = atom_image_pair_matrix_large(9,temp_count_large) + 1.0_dp
             ! Accumulations for the directionally screened fluctuating polarizability
             temp6 = 1.0_dp/(distance**2)
             temp5 = exp(-distance/(total_radius*multibody_screening_parameter))*temp1
             temp4 = temp5/distance**5
             atom_image_pair_matrix_large(11,temp_count_large) = atom_image_pair_matrix_large(11,temp_count_large) &
             + temp4*(distance**2 - 3.0_dp*delta_x**2)
             atom_image_pair_matrix_large(12,temp_count_large) = atom_image_pair_matrix_large(12,temp_count_large) &
             - 3.0_dp*temp4*delta_x*delta_y
             atom_image_pair_matrix_large(13,temp_count_large) = atom_image_pair_matrix_large(13,temp_count_large) &
             - 3.0_dp*temp4*delta_x*delta_z
             atom_image_pair_matrix_large(14,temp_count_large) = atom_image_pair_matrix_large(14,temp_count_large) &
             + temp4*(distance**2 - 3.0_dp*delta_y**2)
             atom_image_pair_matrix_large(15,temp_count_large) = atom_image_pair_matrix_large(15,temp_count_large) &
             - 3.0_dp*temp4*delta_y*delta_z
             atom_image_pair_matrix_large(16,temp_count_large) = atom_image_pair_matrix_large(16,temp_count_large) &
             + temp4*(distance**2 - 3.0_dp*delta_z**2)
             IF (active_pair_small) THEN
               atom_image_pair_matrix_small(1,temp_count_small) = atom1
               atom_image_pair_matrix_small(2,temp_count_small) = atom2
               atom_image_pair_matrix_small(5,temp_count_small) = 1.0_dp - 3.0_dp*temp6*delta_x**2
               atom_image_pair_matrix_small(6,temp_count_small) = -3.0_dp*temp6*delta_x*delta_y
               atom_image_pair_matrix_small(7,temp_count_small) = -3.0_dp*temp6*delta_x*delta_z
               atom_image_pair_matrix_small(8,temp_count_small) = 1.0_dp - 3.0_dp*temp6*delta_y**2
               atom_image_pair_matrix_small(9,temp_count_small) = -3.0_dp*temp6*delta_y*delta_z
               atom_image_pair_matrix_small(10,temp_count_small) = 1.0_dp - 3.0_dp*temp6*delta_z**2
               atom_image_pair_matrix_small(11,temp_count_small) = distance
               atom_image_pair_matrix_small(12,temp_count_small) = temp5
               atom_image_pair_matrix_small(13,temp_count_small) = temp1
               atom_image_pair_matrix_small(14,temp_count_small) = repeata 
               atom_image_pair_matrix_small(15,temp_count_small) = repeatb
               atom_image_pair_matrix_small(16,temp_count_small) = repeatc
             END IF
          END DO
          IF (active_pair_large) THEN
             temp_count_large = temp_count_large + 1
          END IF
       END DO
    END DO
 END DO
!$omp end parallel do

 WRITE(output_FID,*)' '
 WRITE(output_FID,*)'The atom image pair matrices have been initialized.' 
 WRITE(output_FID,*)' '
 FLUSH(output_FID)
 
 IF (num_sym_unique_atom_image_pairs_large < 1001) THEN
   WRITE(output_FID,*)' '
   WRITE(output_FID,*)'The atom image pair matrix large is:' 
   DO i=1,num_sym_unique_atom_image_pairs_large
      WRITE(output_FID,*)(atom_image_pair_matrix_large(j,i),j=1,16)
   END DO
   WRITE(output_FID,*)' '
   FLUSH(output_FID)
   WRITE(output_FID,*)' '
   WRITE(output_FID,*)'The atom image pair matrix small is:' 
   DO i=1,num_sym_unique_atom_image_pairs_small
      WRITE(output_FID,*)(atom_image_pair_matrix_small(j,i),j=1,16)
   END DO
   WRITE(output_FID,*)' '
   FLUSH(output_FID)
 END IF

 DEALLOCATE(region_pairs)
 DEALLOCATE(region_pair_images)
 DEALLOCATE(region_start_end)
 DEALLOCATE(sorted_atoms)
 DEALLOCATE(sorted_coords)
 DEALLOCATE(pair_count_small)
 DEALLOCATE(pair_count_large)
 
 END SUBROUTINE initialize_atom_image_pair_matrices
 
 END MODULE module_initialize_atom_image_pair_matrices
   