 ! Copyright 2019 Thomas A. Manz. 
 ! Software License: CC BY - Creative Commons Attribution

 MODULE module_screening_functions

 USE precise

 IMPLICIT NONE
  
 CONTAINS
  
 FUNCTION calculate_attenuation_length(freq,atomic_polarizability1,atomic_polarizability2,atomic_wp1,atomic_wp2) &
 result(attenuation_length)
 !===================================================================================
 REAL(kind=dp) :: freq,atomic_polarizability1,atomic_polarizability2,atomic_wp1,atomic_wp2
 REAL(kind=dp) :: attenuation_length

 attenuation_length=sqrt((calculate_frequency_dependent_effective_atomic_radius(freq,atomic_polarizability1,atomic_wp1))**2+&
 (calculate_frequency_dependent_effective_atomic_radius(freq,atomic_polarizability2,atomic_wp2))**2)

 END FUNCTION calculate_attenuation_length
 
 
 
 FUNCTION calculate_frequency_dependent_effective_atomic_radius(freq,static_atomic_polarizability,atomic_wp) &
 result(frequency_dependent_effective_atomic_radius)
 !===================================================================================
 REAL(kind=dp) :: freq,static_atomic_polarizability,atomic_wp,pi = 3.14159265358979323846_dp
 REAL(kind=dp) :: frequency_dependent_effective_atomic_radius

 frequency_dependent_effective_atomic_radius = (sqrt(2.0_dp/pi)*calculate_frequency_dependent_unscreened_atomic_polarizability&
 (freq,static_atomic_polarizability,atomic_wp)/3.0_dp)**(1.0_dp/3.0_dp)
 
 END FUNCTION calculate_frequency_dependent_effective_atomic_radius
  
 

 FUNCTION calculate_frequency_dependent_unscreened_atomic_polarizability(freq,static_unscreened_atomic_polarizability,&
 atomic_wp) &
 result (frequency_dependent_unscreened_atomic_polarizability)
 !===================================================================================
 REAL(kind=dp) :: freq,static_unscreened_atomic_polarizability,atomic_wp
 REAL(kind=dp) :: frequency_dependent_unscreened_atomic_polarizability
 
 frequency_dependent_unscreened_atomic_polarizability = static_unscreened_atomic_polarizability/(1.0_dp+(freq/atomic_wp)**2)

 END FUNCTION calculate_frequency_dependent_unscreened_atomic_polarizability



 FUNCTION smoothed_minimum_function(value1,value2) result(value)
 !--------------------------------------------------------------
 INTEGER :: P
 REAL(kind=dp) :: zero_tolerance
 REAL(kind=dp) :: value1,value2
 REAL(kind=dp) :: value

 P=25
 zero_tolerance = 1.0e-10_dp
 value = 0.0_dp
 IF ((value1 < zero_tolerance) .or. (value2 < zero_tolerance)) THEN
   value = 0.0_dp
 ELSE IF (min(value1,value2)/max(value1,value2) < 1.0e-5_dp) THEN
   value = min(value1,value2)
 ELSE    
   value = sqrt(value1*value2)*(1.0_dp/((value1/value2)**P+(value2/value1)**P-1.0_dp))**(1.0_dp/(2.0_dp*P))
 END IF
 
 END FUNCTION smoothed_minimum_function



 FUNCTION smooth_cutoff_function(value1,value2) result(value)
 !-------------------------------------------------------------- 
 REAL(kind=dp) :: value1,value2
 REAL(kind=dp) :: value
 IF (value1 < value2) THEN
    value = 1.0_dp - exp(-20.0_dp*(1.0_dp - value1/value2)**3)
 ELSE
    value = 0.0_dp
 END IF
 
 END FUNCTION smooth_cutoff_function

 
 END MODULE module_screening_functions